# This file is part of Invenio.
# Copyright (C) 2007, 2008, 2009, 2010, 2011, 2013, 2014, 2015, 2016 CERN.
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

# cython: infer_types=True
# cython: language_level=2

"""
Defines an intbitset data object to hold unordered sets of unsigned
integers with ultra fast set operations, implemented via bit vectors
and Python C extension to optimize speed and memory usage.

Emulates the Python built-in set class interface with some additional
specific methods such as its own fast dump and load marshalling
functions.  Uses real bits to optimize memory usage, so may have
issues with endianness if you transport serialized bitsets between
various machine architectures.

Please note that no bigger than __maxelem__ elements can be added to
an intbitset and, if CFG_INTBITSET_ENABLE_SANITY_CHECKS is disabled,
you will receive unpredictable results.

Note to developers: If you make modification to this file you
have to manually regenerate intbitset.c by running:
  $ cython intbitset.pyx
and then commit generated intbitset.c.
"""

#cython: infer_types=True

import sys
import zlib
from array import array
from cpython.buffer cimport PyBUF_SIMPLE, Py_buffer, PyObject_GetBuffer, PyBuffer_Release
from cpython.ref cimport PyObject

CFG_INTBITSET_ENABLE_SANITY_CHECKS = False
from intbitset_helper import _
from intbitset_version import __version__

__all__ = ['intbitset', '__version__']

cdef extern from *:
    ## See: <http://wiki.cython.org/FAQ/#HowdoIuse.27const.27.3F>
    ## In order to avoid warnings with PyObject_AsReadBuffer
    ctypedef void* const_void_ptr "const void*"

cdef extern from "intbitset.h":
    ctypedef int Py_ssize_t
    object PyBytes_FromStringAndSize(char *s, Py_ssize_t len)
    object PyString_FromStringAndSize(char *s, Py_ssize_t len)

cdef extern from "intbitset.h":
    ctypedef unsigned long long int word_t
    #ctypedef unsigned char bool_t
    ctypedef struct IntBitSet:
        int size
        int allocated
        word_t trailing_bits
        int tot
        word_t *bitset
    int wordbytesize
    int wordbitsize
    int maxelem
    IntBitSet *intBitSetCreate(int size, bint trailing_bits)
    IntBitSet *intBitSetCreateFromBuffer(void *buf, int bufsize)
    IntBitSet *intBitSetResetFromBuffer(IntBitSet *bitset, void *buf, int bufsize)
    IntBitSet *intBitSetReset(IntBitSet *bitset)
    void intBitSetDestroy(IntBitSet *bitset)
    IntBitSet *intBitSetClone(IntBitSet *bitset)
    int intBitSetGetSize(IntBitSet *bitset)
    int intBitSetGetAllocated(IntBitSet *bitset)
    int intBitSetGetTot(IntBitSet * bitset)
    bint intBitSetIsInElem(IntBitSet *bitset, unsigned int elem)
    void intBitSetAddElem(IntBitSet *bitset, unsigned int elem)
    void intBitSetDelElem(IntBitSet *bitset, unsigned int elem)
    bint intBitSetEmpty(IntBitSet *bitset)
    IntBitSet *intBitSetUnion(IntBitSet *x, IntBitSet *y)
    IntBitSet *intBitSetIntersection(IntBitSet *x, IntBitSet *y)
    IntBitSet *intBitSetSub(IntBitSet *x, IntBitSet *y)
    IntBitSet *intBitSetXor(IntBitSet *x, IntBitSet *y)
    IntBitSet *intBitSetIUnion(IntBitSet *dst, IntBitSet *src)
    IntBitSet *intBitSetIIntersection(IntBitSet *dst, IntBitSet *src)
    IntBitSet *intBitSetISub(IntBitSet *x, IntBitSet *y)
    IntBitSet *intBitSetIXor(IntBitSet *x, IntBitSet *y)
    bint intBitSetIsDisjoint(IntBitSet *x, IntBitSet *y)
    int intBitSetGetNext(IntBitSet *x, int last)
    int intBitSetGetLast(IntBitSet *x)
    unsigned char intBitSetCmp(IntBitSet *x, IntBitSet *y)

__maxelem__ = maxelem

cdef class intbitset:
    """
    Defines an intbitset data object to hold unordered sets of
    unsigned integers with ultra fast set operations, implemented via
    bit vectors and Python C extension to optimize speed and memory
    usage.

    Emulates the Python built-in set class interface with some
    additional specific methods such as its own fast dump and load
    marshalling functions.  Uses real bits to optimize memory usage,
    so may have issues with endianness if you transport serialized
    bitsets between various machine architectures.

    The constructor accept the following parameters:
        rhs=0, 
        int preallocate=-1, 
        int trailing_bits=0,
        bint sanity_checks=CFG_INTBITSET_ENABLE_SANITY_CHECKS,
        int no_allocate=0:

    where rhs can be:

    * ``int/long`` for creating allocating empty intbitset that will hold at
      least rhs elements, before being resized

    * ``intbitset`` for cloning

    * ``bytes`` for retrieving an intbitset that was dumped into a byte string

    * ``array`` for retrieving an intbitset that was dumped into a string stored
      in an array

    * sequence made of integers for copying all the elements from the sequence. 
      If minsize is specified than it is initially allocated enough space to
      hold up to minsize integers, otherwise the biggest element of the sequence
      will be used.

    * sequence made of tuples: then the first element of each tuple is
      considered as an integer (as in the sequence made of integers).

    The other arguments are:

    * ``preallocate`` is a suggested initial upper bound on the numbers that
      will be stored, by looking at rhs a sequence of number.

    * ``trailing_bits`` is 1, then the set will contain "all" the positive integers

    * ``no_allocate`` and ``sanity_checks`` are used internally and should never be set.
    """
    cdef IntBitSet *bitset
    cdef bint sanity_checks

    def __cinit__(
        self not None, 
        rhs=0, 
        int preallocate=-1, 
        int trailing_bits=0, 
        bint sanity_checks=CFG_INTBITSET_ENABLE_SANITY_CHECKS, 
        int no_allocate=0,
    ):
        cdef Py_ssize_t size = 0
        cdef const_void_ptr buf = NULL
        cdef int elem
        cdef int last
        cdef int remelem
        cdef bint tuple_of_tuples
        cdef Py_buffer view

        self.sanity_checks = sanity_checks
        #print >> sys.stderr, "intbitset.__cinit__ is called"
        msg = "Error"
        self.bitset = NULL
        try:
            if no_allocate:
                return
            if type(rhs) in (int, long):
                if rhs < 0:
                    raise ValueError("rhs can't be negative")
                self.bitset = intBitSetCreate(rhs, trailing_bits)
            elif type(rhs) is intbitset:
                self.bitset = intBitSetClone((<intbitset>rhs).bitset)
            elif type(rhs) in (bytes, array):
                try:
                    if type(rhs) is array:
                        rhs = rhs.tobytes()
                    tmp = zlib.decompress(rhs)

                    if PyObject_GetBuffer(tmp, &view, PyBUF_SIMPLE) != 0:
                        raise ValueError("Unable to get buffer")
                    
                    try:
                        buf = <const_void_ptr>view.buf
                        size = view.len

                        if (size % wordbytesize):
                            ## Wrong size!
                            raise Exception()

                        self.bitset = intBitSetCreateFromBuffer(buf, size)
                    finally:
                        PyBuffer_Release(&view)
                        
                except Exception as e:
                    raise ValueError("rhs is corrupted: %s" % str(e))
            elif hasattr(rhs, '__iter__'):
                tuple_of_tuples = (
                    rhs
                    and hasattr(rhs, '__getitem__')
                    and hasattr(rhs[0], '__getitem__')
                )
                try:
                    if preallocate < 0:
                        if rhs and (not hasattr(rhs, '__getitem__') or type(rhs[0]) is int):
                            try:
                                preallocate = max(rhs)
                            except ValueError:
                                preallocate = 0
                        else:
                            preallocate = 0
                    if self.sanity_checks:
                        if not (0 <= preallocate < maxelem):
                            raise OverflowError("Can't store integers bigger than %s" % maxelem)
                    self.bitset = intBitSetCreate(preallocate, trailing_bits)
                    if trailing_bits:
                        last = 0
                        if self.sanity_checks:
                            if tuple_of_tuples:
                                for tmp_tuple in rhs:
                                    elem = tmp_tuple[0]
                                    if elem < 0:
                                        raise ValueError("Negative numbers, not allowed")
                                    elif elem > maxelem:
                                        raise OverflowError("Elements must be <= %s" % maxelem)
                                    for remelem from last <= remelem < elem:
                                        intBitSetDelElem(self.bitset, remelem)
                                    last = elem + 1
                            else:
                                for elem in rhs:
                                    if elem < 0:
                                        raise ValueError("Negative numbers, not allowed")
                                    elif elem > maxelem:
                                        raise OverflowError("Elements must be <= %s" % maxelem)
                                    for remelem from last <= remelem < elem:
                                        intBitSetDelElem(self.bitset, remelem)
                                    last = elem + 1
                        else:
                            if tuple_of_tuples:
                                for tmp_tuple in rhs:
                                    elem = tmp_tuple[0]
                                    for remelem from last <= remelem < elem:
                                        intBitSetDelElem(self.bitset, remelem)
                                    last = elem + 1
                            else:
                                for elem in rhs:
                                    for remelem from last <= remelem < elem:
                                        intBitSetDelElem(self.bitset, remelem)
                                    last = elem + 1

                    else:
                        if self.sanity_checks:
                            if tuple_of_tuples:
                                for tmp_tuple in rhs:
                                    elem = tmp_tuple[0]
                                    if elem < 0:
                                        raise ValueError("Negative numbers, not allowed")
                                    elif elem > maxelem:
                                        raise OverflowError("Elements must be <= %s" % maxelem)
                                    intBitSetAddElem(self.bitset, elem)
                            else:
                                for elem in rhs:
                                    if elem < 0:
                                        raise ValueError("Negative numbers, not allowed")
                                    elif elem > maxelem:
                                        raise OverflowError("Elements must be <= %s" % maxelem)
                                    intBitSetAddElem(self.bitset, elem)
                        else:
                            if tuple_of_tuples:
                                for tmp_tuple in rhs:
                                    elem = tmp_tuple[0]
                                    intBitSetAddElem(self.bitset, elem)
                            else:
                                for elem in rhs:
                                    intBitSetAddElem(self.bitset, elem)
                except Exception as e:
                    raise ValueError("retrieving integers from rhs is impossible: %s" % str(e))
            else:
                raise TypeError("rhs is of unknown type %s" % type(rhs))
        except:
            intBitSetDestroy(self.bitset)
            raise

    def __dealloc__(self not None):
        #print >> sys.stderr, "intbitset.__dealloc__ is called"
        intBitSetDestroy(self.bitset)

    def __contains__(self not None, int elem):
        if self.sanity_checks:
            if elem < 0:
                raise ValueError("Negative numbers, not allowed")
            elif elem > maxelem:
                raise OverflowError("Element must be <= %s" % maxelem)
        return intBitSetIsInElem(self.bitset, elem) != 0

    def __cmp__(self not None, intbitset rhs not None):
        raise TypeError("cannot compare intbitset using cmp()")

    def __richcmp__(self not None, rhs, int op):
        if not isinstance(self, intbitset) or not isinstance(rhs, intbitset):
            return False
        cdef short unsigned int tmp
        tmp = intBitSetCmp((<intbitset>self).bitset, (<intbitset>rhs).bitset)
        if op == 0: # <
            return tmp == 1
        if op == 1: # <=
            return tmp <= 1
        if op == 2: # ==
            return tmp == 0
        if op == 3: # !=
            return tmp > 0
        if op == 4: # >
            return tmp == 2
        if op == 5: # >=
            return tmp in (0, 2)

    def __len__(self not None):
        return intBitSetGetTot(self.bitset)

    def __hash__(self not None):
        return hash(
            PyBytes_FromStringAndSize(
                <char *>self.bitset.bitset, 
                wordbytesize * (intBitSetGetTot(self.bitset) / wordbitsize + 1)
            )
        )

    def __nonzero__(self not None):
        return not intBitSetEmpty(self.bitset)

    def __deepcopy__(self not None, memo):
        return intbitset(self)

    def __delitem__(self not None, int elem):
        if self.sanity_checks:
            if elem < 0:
                raise ValueError("Negative numbers, not allowed")
            elif elem > maxelem:
                raise OverflowError("Element must be <= %s" % maxelem)
        intBitSetDelElem(self.bitset, elem)

    def __iadd__(self not None, rhs):
        cdef int elem
        if isinstance(rhs, (int, long)):
            if self.sanity_checks:
                if rhs < 0:
                    raise ValueError("Negative numbers, not allowed")
                elif rhs > maxelem:
                    raise OverflowError("rhs must be <= %s" % maxelem)
            intBitSetAddElem(self.bitset, rhs)
        elif isinstance(rhs, intbitset):
            intBitSetIUnion(self.bitset, (<intbitset> rhs).bitset)
        else:
            if self.sanity_checks:
                for elem in rhs:
                    if elem < 0:
                        raise ValueError("Negative numbers, not allowed")
                    elif elem > maxelem:
                        raise OverflowError("Elements must be <= %s" % maxelem)
                    intBitSetAddElem(self.bitset, elem)
            else:
                for elem in rhs:
                    if elem < 0:
                        raise ValueError("Negative numbers, not allowed")
                    elif elem > maxelem:
                        raise OverflowError("Elements must be <= %s" % maxelem)
                    intBitSetAddElem(self.bitset, elem)
        return self

    def __isub__(self not None, rhs not None):
        """Remove all elements of another set from this set."""
        cdef int elem
        if isinstance(rhs, (int, long)):
            if self.sanity_checks:
                if rhs < 0:
                    raise ValueError("Negative numbers, not allowed")
                elif rhs > maxelem:
                    raise OverflowError("rhs must be <= %s" % maxelem)
            intBitSetDelElem(self.bitset, rhs)
        elif isinstance(rhs, intbitset):
            intBitSetISub(self.bitset, (<intbitset> rhs).bitset)
        else:
            if self.sanity_checks:
                for elem in rhs:
                    if elem < 0:
                        raise ValueError("Negative numbers, not allowed")
                    elif elem > maxelem:
                        raise OverflowError("Elements must be <= %s" % maxelem)
                    intBitSetDelElem(self.bitset, elem)
            else:
                for elem in rhs:
                    intBitSetDelElem(self.bitset, elem)
        return self

    def __sub__(self not None, intbitset rhs not None):
        """Return the difference of two intbitsets as a new set.
        (i.e. all elements that are in this intbitset but not the other.)
        """
        cdef intbitset ret = intbitset(no_allocate=1)
        (<intbitset>ret).bitset = intBitSetSub((<intbitset> self).bitset, rhs.bitset)
        return ret

    def __and__(self not None, intbitset rhs not None):
        """Return the intersection of two intbitsets as a new set.
        (i.e. all elements that are in both intbitsets.)
        """
        cdef intbitset ret = intbitset(no_allocate=1)
        (<intbitset>ret).bitset = intBitSetIntersection((<intbitset> self).bitset, rhs.bitset)
        return ret

    def __iand__(self not None, intbitset rhs not None):
        """Update a intbitset with the intersection of itself and another."""
        intBitSetIIntersection(self.bitset, rhs.bitset)
        return self

    def __or__(self not None, intbitset rhs not None):
        """Return the union of two intbitsets as a new set.
        (i.e. all elements that are in either intbitsets.)
        """
        cdef intbitset ret = intbitset(no_allocate=1)
        (<intbitset>ret).bitset = intBitSetUnion((<intbitset> self).bitset, rhs.bitset)
        return ret

    def __ior__(self not None, intbitset rhs not None):
        """Update a intbitset with the union of itself and another."""
        intBitSetIUnion(self.bitset, rhs.bitset)
        return self

    def __xor__(self not None, intbitset rhs not None):
        """Return the symmetric difference of two sets as a new set.
        (i.e. all elements that are in exactly one of the sets.)
        """
        cdef intbitset ret = intbitset(no_allocate=1)
        (<intbitset>ret).bitset = intBitSetXor((<intbitset> self).bitset, rhs.bitset)
        return ret

    def __ixor__(self not None, intbitset rhs not None):
        """Update an intbitset with the symmetric difference of itself and another.
        """
        intBitSetIXor(self.bitset, rhs.bitset)
        return self

    def __repr__(self not None):
        finite_list = self.extract_finite_list()
        if self.bitset.trailing_bits:
            return "intbitset(%s, trailing_bits=True)" % repr(finite_list)
        else:
            return "intbitset(%s)" % repr(finite_list)

    def __str__(self not None):
        cdef int tot
        tot = intBitSetGetTot(self.bitset)
        if tot < 0:
            return "intbitset([...], trailing_bits=True)"
        elif tot > 10:
            begin_list = self[0:5]
            end_list = self[tot - 5:tot]
            ret = "intbitset(["
            for n in begin_list:
                ret += '%i, ' % n
            ret += "..., "
            for n in end_list:
                ret += '%i, ' % n
            ret = ret[:-2]
            ret += '])'
            return ret
        else:
            return self.__repr__()

    def __getitem__(self not None, object key):
        cdef Py_ssize_t i
        cdef int elem = -1
        cdef int start
        cdef int end
        cdef int step
        if hasattr(key, 'indices'):
            ## This is a slice object!
            if self.bitset.trailing_bits and (key.start < 0 or key.stop < 0):
                raise IndexError("negative indexes are not allowed on infinite intbitset")
            retset = intbitset()
            start, end, step = key.indices(intBitSetGetTot(self.bitset))
            if step < 0:
                raise ValueError("negative steps are not yet supported")
            for i in range(start):
                elem = intBitSetGetNext(self.bitset, elem)
                if elem < 0:
                    return retset
            for i in range(end - start):
                elem = intBitSetGetNext(self.bitset, elem)
                if elem < 0:
                    return retset
                if i % step == 0:
                    retset.add(elem)
            return retset
        else:
            end = key
            if end < 0:
                if self.bitset.trailing_bits:
                    raise IndexError("negative indexes are not allowed on infinite intbitset")
                end += intBitSetGetTot(self.bitset)
                if end < 0:
                    raise IndexError("intbitset index out of range")
            if end >= intBitSetGetTot(self.bitset):
                raise IndexError("intbitset index out of range")
            for i in range(end + 1):
                elem = intBitSetGetNext(self.bitset, elem)
            return elem

    # pickle interface
    def __reduce__(self not None):
        return _, (self.fastdump(),)

    __safe_for_unpickling__ = True

    # Iterator interface
    def __iter__(self not None):
        if self.bitset.trailing_bits:
            raise OverflowError("It's impossible to iterate over an infinite set.")
        return intbitset_iterator(self)

    # Customized interface
    cpdef add(intbitset self, int elem):
        """Add an element to a set.
        This has no effect if the element is already present."""
        if self.sanity_checks:
            if elem < 0:
                raise ValueError("Negative numbers, not allowed")
            elif elem > maxelem:
                raise OverflowError("Element must be <= %s" % maxelem)
        intBitSetAddElem(self.bitset, elem)

    cpdef clear(intbitset self):
        intBitSetReset(self.bitset)


    cpdef discard(intbitset self, int elem):
        """Remove an element from a intbitset if it is a member.
        If the element is not a member, do nothing."""
        if self.sanity_checks:
            if elem < 0:
                raise ValueError("Negative numbers, not allowed")
            elif elem > maxelem:
                raise OverflowError("Element must be <= %s" % maxelem)
        intBitSetDelElem(self.bitset, elem)

    symmetric_difference = __xor__
    symmetric_difference_update = __ixor__

    cpdef issubset(intbitset self, rhs):
        """Report whether another set contains this set."""
        return self.__le__(rhs)

    cpdef issuperset(intbitset self, rhs):
        """Report whether this set contains another set."""
        return self.__ge__(rhs)

    # Dumping & Loading
    cpdef fastdump(intbitset self):
        """Return a compressed string representation suitable to be saved
        somewhere."""
        cdef Py_ssize_t size
        size = intBitSetGetSize((<intbitset> self).bitset)
        tmp = PyBytes_FromStringAndSize(<char *>self.bitset.bitset, ( size + 1) * wordbytesize)
        return zlib.compress(tmp)

    cpdef fastload(intbitset self, strdump):
        """Load a compressed string representation produced by a previous call
        to the fastdump method into the current intbitset. The previous content
        will be replaced."""
        cdef Py_ssize_t size
        cdef const_void_ptr buf
        cdef Py_buffer view

        buf = NULL
        size = 0
        try:
            if type(strdump) is array:
                strdump = strdump.tostring()
            # tmp needed to not be garbage collected
            tmp = zlib.decompress(strdump)

            if PyObject_GetBuffer(tmp, &view, PyBUF_SIMPLE) != 0:
                raise ValueError("Unable to get buffer")

            try:
                buf = <const_void_ptr>view.buf
                size = view.len

                if (size % wordbytesize):
                    ## Wrong size!
                    raise Exception()

                intBitSetResetFromBuffer((<intbitset> self).bitset, buf, size)
            finally:
                PyBuffer_Release(&view)

        except:
            raise ValueError("strdump is corrupted")

    cpdef copy(intbitset self):
        """Return a shallow copy of a set."""
        return intbitset(self)

    cpdef pop(intbitset self):
        """Remove and return an arbitrary set element.

        Note: intbitset implementation of .pop() differs from the native ``set``
            implementation by guaranteeing returning always the largest element.
        """
        cdef int ret
        ret = intBitSetGetLast(self.bitset)
        if ret < 0:
            raise KeyError("pop from an empty or infinite intbitset")
        intBitSetDelElem(self.bitset, ret)
        return ret

    cpdef remove(intbitset self, int elem):
        """Remove an element from a set; it must be a member.
        If the element is not a member, raise a KeyError.
        """
        if self.sanity_checks:
            if elem < 0:
                raise ValueError("Negative numbers, not allowed")
            elif elem > maxelem:
                raise OverflowError("Elements must be <= %s" % maxelem)
        if intBitSetIsInElem(self.bitset, elem):
            intBitSetDelElem(self.bitset, elem)
        else:
            raise KeyError(elem)

    cpdef strbits(intbitset self):
        """Return a string of 0s and 1s representing the content in memory
        of the intbitset.
        """
        cdef int i
        cdef int last
        if (<intbitset> self).bitset.trailing_bits:
            raise OverflowError("It's impossible to print an infinite set.")
        last = 0
        ret = []
        for i in self:
            ret.append('0'*(i-last)+'1')
            last = i+1
        return ''.join(ret)

    def update(self not None, *args):
        """Update the intbitset, adding elements from all others."""
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetIUnion(self.bitset, iarg.bitset)

    union_update = update

    def intersection_update(self not None, *args):
        """Update the intbitset, keeping only elements found in it and all others."""
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetIIntersection(self.bitset, iarg.bitset)

    def difference_update(self not None, *args):
        """Update the intbitset, removing elements found in others."""
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetISub(self.bitset, iarg.bitset)

    def union(self not None, *args):
        """Return a new intbitset with elements from the intbitset and all others."""
        cdef intbitset ret = intbitset(self)
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetIUnion(ret.bitset, iarg.bitset)
        return ret

    def intersection(self not None, *args):
        """Return a new intbitset with elements common to the intbitset and all others."""
        cdef intbitset ret = intbitset(self)
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetIIntersection(ret.bitset, iarg.bitset)
        return ret

    def difference(self not None, *args):
        """Return a new intbitset with elements from the intbitset that are not in the others."""
        cdef intbitset ret = intbitset(self)
        cdef intbitset iarg
        for arg in args:
            iarg = arg if hasattr(arg, "bitset") else intbitset(arg)
            intBitSetISub(ret.bitset, iarg.bitset)
        return ret

    def isdisjoint(self not None, intbitset rhs not None):
        """Return True if two intbitsets have a null intersection."""
        return intBitSetIsDisjoint(self.bitset, rhs.bitset)

    cpdef update_with_signs(intbitset self, rhs):
        """Given a dictionary rhs whose keys are integers, remove all the integers
        whose value are less than 0 and add every integer whose value is 0 or more"""
        cdef int value
        try:
            if self.sanity_checks:
                for value, sign in rhs.iteritems():
                    if value < 0:
                        raise ValueError("Negative numbers, not allowed")
                    elif value > maxelem:
                        raise OverflowError("Elements must <= %s" % maxelem)
                    if sign < 0:
                        intBitSetDelElem(self.bitset, value)
                    else:
                        intBitSetAddElem(self.bitset, value)
            else:
                for value, sign in rhs.iteritems():
                    if sign < 0:
                        intBitSetDelElem(self.bitset, value)
                    else:
                        intBitSetAddElem(self.bitset, value)
        except AttributeError:
            raise TypeError("rhs should be a valid dictionary with integers keys and integer values")

    cpdef get_size(intbitset self):
        return intBitSetGetSize(self.bitset)

    cpdef get_allocated(intbitset self):
        return intBitSetGetAllocated(self.bitset)

    cpdef is_infinite(intbitset self):
        """Return True if the intbitset is infinite. (i.e. trailing_bits=True
        was used in the constructor.)"""
        return self.bitset.trailing_bits != 0

    cpdef extract_finite_list(intbitset self, int up_to=-1):
        """Return a finite list of elements sufficient to be passed to intbitset
        constructor toghether with the proper value of trailing_bits in order
        to reproduce this intbitset. At least up_to integer are looked for when
        they are inside the intbitset but not necessarily needed to build the
        intbitset"""
        cdef int true_up_to
        cdef int last
        if self.sanity_checks and up_to > maxelem:
            raise OverflowError("up_to must be <= %s" % maxelem)
        ret = []
        true_up_to = max(up_to, (intBitSetGetSize(self.bitset)) * wordbitsize)
        last = -1
        while last < true_up_to:
            last = intBitSetGetNext(self.bitset, last)
            if last == -2:
                break
            ret.append(last)
        return ret

    cpdef get_wordbitsize(intbitset self):
        return wordbitsize

    cpdef get_wordbytsize(intbitset self):
        return wordbytesize

    cpdef tolist(intbitset self):
        """Legacy method to retrieve a list of all the elements inside an
        intbitset.
        """
        if self.bitset.trailing_bits:
            raise OverflowError("It's impossible to retrieve a list of an infinite set")
        return self.extract_finite_list()

    cdef object __weakref__

cdef class intbitset_iterator:
    cdef int last
    cdef IntBitSet *bitset
    cdef bint sanity_checks

    def __cinit__(intbitset_iterator self, intbitset bitset not None):
        #print >> sys.stderr, "intbitset_iterator.__cinit__ is called"
        self.last = -1
        ## A copy should be performed, in case the original bitset disappears
        ## as in "for x in intbitset([1,2,3])"!
        self.bitset = intBitSetClone(bitset.bitset)
        self.sanity_checks = CFG_INTBITSET_ENABLE_SANITY_CHECKS

    def __dealloc__(intbitset_iterator self):
        #print >> sys.stderr, "intbitset_iterator.__dealloc__ is called"
        intBitSetDestroy(self.bitset)

    def __next__(intbitset_iterator self):
        if self.last == -2:
            raise StopIteration()
        self.last = intBitSetGetNext(self.bitset, self.last)
        if self.sanity_checks and (self.bitset.allocated < self.bitset.size):
            raise MemoryError(
                "intbitset corrupted: allocated: %s, size: %s"
                % (self.bitset.allocated, self.bitset.size)
            )
        if self.last < 0:
            self.last = -2
            raise StopIteration()
        return self.last

    def __iter__(intbitset_iterator self):
        return self

    cdef object __weakref__



# -*- coding: utf-8 -*-
#
# This file is part of intbitset.
# Copyright (C) 2007, 2008, 2009, 2010, 2011, 2013, 2014, 2015, 2016 CERN.
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# intbitset is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# intbitset is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with intbitset; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

import copy
import pickle
import zlib
from typing import NamedTuple

import pytest

from intbitset import intbitset


def b(s):
    # carried from six
    return s.encode("latin-1")


def check_bitset(intbitset1, msg=""):
    """
    Run basic sanity checks on a bitset
    """
    wordbitsize = intbitset1.get_wordbitsize()
    size1 = intbitset1.get_size()
    allocated1 = intbitset1.get_allocated()
    creator_list = intbitset1.extract_finite_list()
    up_to1 = creator_list and max(creator_list) or -1
    assert up_to1 <= size1 * wordbitsize < allocated1 * wordbitsize, (
        "up_to1=%s, size1=%s, allocated1=%s while testing %s during %s"
        % (up_to1, size1 * wordbitsize, allocated1 * wordbitsize, intbitset1, msg),
    )
    tmp = intbitset(intbitset1.fastdump())
    size2 = tmp.get_size()
    allocated2 = tmp.get_allocated()
    creator_list = tmp.extract_finite_list()
    up_to2 = creator_list and max(creator_list) or -1
    assert up_to2 <= size2 * wordbitsize < allocated2 * wordbitsize, (
        "After serialization up_to2=%s, size2=%s, allocated2=%s while testing %s during %s"
        % (up_to2, size2 * wordbitsize, allocated2 * wordbitsize, intbitset1, msg),
    )


TEST_SETS = [
    [1024],
    [10, 20],
    [10, 40],
    [60, 70],
    [60, 80],
    [10, 20, 60, 70],
    [10, 40, 60, 80],
    [1000],
    [10000],
    [23, 45, 67, 89, 110, 130, 174, 1002, 2132, 23434],
    [700, 2000],
    list(range(1000, 1100)),
    [30],
    [31],
    [32],
    [33],
    [62],
    [63],
    [64],
    [65],
    [126],
    [127],
    [128],
    [129],
]


class Function(NamedTuple):
    intbitset_function: str
    set_function: str
    int_function: str
    label: str
    inplace: bool


AND_FUNC = Function(
    intbitset_function=intbitset.__and__,  # NOQA
    set_function=set.__and__,
    int_function=int.__and__,
    label="intersection",
    inplace=False,
)

IAND_FUNC = Function(
    intbitset_function=intbitset.__iand__,  # NOQA
    set_function=set.__iand__,
    int_function=int.__and__,
    label="intersection_update",
    inplace=True,
)

OR_FUNC = Function(
    intbitset_function=intbitset.__or__,  # NOQA
    set_function=set.__or__,
    int_function=int.__or__,
    label="union",
    inplace=False,
)

IOR_FUNC = Function(
    intbitset_function=intbitset.__ior__,  # NOQA
    set_function=set.__ior__,
    int_function=int.__or__,
    label="union_update",
    inplace=True,
)

XOR_FUNC = Function(
    intbitset_function=intbitset.__xor__,  # NOQA
    set_function=set.__xor__,
    int_function=int.__xor__,
    label="symmetric_difference",
    inplace=False,
)

IXOR_FUNC = Function(
    intbitset_function=intbitset.__ixor__,  # NOQA
    set_function=set.__ixor__,
    int_function=int.__xor__,
    label="symmetric_difference_update",
    inplace=True,
)

SUB_FUNC = Function(
    intbitset_function=intbitset.__sub__,  # NOQA
    set_function=set.__sub__,
    int_function=int.__sub__,
    label="difference",
    inplace=False,
)

ISUB_FUNC = Function(
    intbitset_function=intbitset.__isub__,  # NOQA
    set_function=set.__isub__,
    int_function=int.__sub__,
    label="difference_update",
    inplace=True,
)

FUNCTIONS = [
    SUB_FUNC,
    ISUB_FUNC,
    AND_FUNC,
    IAND_FUNC,
    OR_FUNC,
    IOR_FUNC,
    XOR_FUNC,
    IXOR_FUNC,
]


def check_functions_work(function, intbitset1, intbitset2):
    function = function
    orig1 = intbitset(intbitset1)
    orig2 = intbitset(intbitset2)

    msg = "Testing %s(%s, %s)" % (
        function.intbitset_function.__name__,
        repr(intbitset1),
        repr(intbitset2),
    )

    trailing1 = intbitset1.is_infinite()
    trailing2 = intbitset2.is_infinite()

    if function.inplace:
        function.intbitset_function(intbitset1, intbitset2)
        trailing1 = function.int_function(trailing1, trailing2) > 0
        up_to = intbitset1.extract_finite_list() and max(intbitset1.extract_finite_list()) or -1
    else:
        intbitset3 = function.intbitset_function(intbitset1, intbitset2)
        trailing3 = function.int_function(trailing1, trailing2) > 0
        up_to = intbitset3.extract_finite_list() and max(intbitset3.extract_finite_list()) or -1

    set1 = set(orig1.extract_finite_list(up_to))
    set2 = set(orig2.extract_finite_list(up_to))

    if function.inplace:
        function.set_function(set1, set2)
    else:
        set3 = function.set_function(set1, set2)

    check_bitset(intbitset1, msg)
    check_bitset(intbitset2, msg)

    if function.inplace:
        assert set1 & set(intbitset1.extract_finite_list(up_to)) == set(
            intbitset1.extract_finite_list(up_to)
        ), (
            "%s not equal to %s after executing %s(%s, %s)"
            % (
                set1,
                set(intbitset1.extract_finite_list(up_to)),
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )
        assert set1 | set(intbitset1.extract_finite_list(up_to)) == set1, (
            "%s not equal to %s after executing %s(%s, %s)"
            % (
                set1,
                set(intbitset1.extract_finite_list(up_to)),
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )
        assert trailing1 == intbitset1.is_infinite(), (
            "%s is not %s as it is supposed to be after executing %s(%s, %s)"
            % (
                intbitset1,
                trailing1 and "infinite" or "finite",
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )
    else:
        check_bitset(intbitset3, msg)
        assert set3 & set(intbitset3.extract_finite_list(up_to)) == set(
            intbitset3.extract_finite_list(up_to)
        ), (
            "%s not equal to %s after executing %s(%s, %s)"
            % (
                set3,
                set(intbitset3.extract_finite_list(up_to)),
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )
        assert set3 | set(intbitset3.extract_finite_list(up_to)) == set3, (
            "%s not equal to %s after executing %s(%s, %s)"
            % (
                set3,
                set(intbitset3.extract_finite_list(up_to)),
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )
        assert trailing3 == intbitset3.is_infinite(), (
            "%s is not %s as it is supposed to be after executing %s(%s, %s)"
            % (
                intbitset3,
                trailing3 and "infinite" or "finite",
                function.intbitset_function.__name__,
                repr(orig1),
                repr(orig2),
            ),
        )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS,
)
@pytest.mark.parametrize(argnames="set2", argvalues=TEST_SETS)
@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_normal_set_ops(set1, set2, function):
    check_functions_work(function, intbitset(set1), intbitset(set2))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS,
)
@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_empty_set(set1, function):
    check_functions_work(function, intbitset(set1), intbitset([]))
    check_functions_work(function, intbitset([]), intbitset(set1))


@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_empty_empty_set(function):
    check_functions_work(function, intbitset([]), intbitset([]))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS,
)
@pytest.mark.parametrize(argnames="set2", argvalues=TEST_SETS)
@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_inifinite_set(set1, set2, function):

    check_functions_work(function, intbitset(set1), intbitset(set2, trailing_bits=True))

    check_functions_work(function, intbitset(set1, trailing_bits=True), intbitset(set2))

    check_functions_work(
        function,
        intbitset(set1, trailing_bits=True),
        intbitset(set2, trailing_bits=True),
    )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS,
)
@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_infinite_vs_empty_set(set1, function):
    check_functions_work(function, intbitset(set1, trailing_bits=True), intbitset([]))
    check_functions_work(function, intbitset([]), intbitset(set1, trailing_bits=True))


@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_infinite_empty_empty_set(function):
    check_functions_work(function, intbitset([]), intbitset(trailing_bits=True))
    check_functions_work(function, intbitset(trailing_bits=True), intbitset([]))


@pytest.mark.parametrize(
    argnames="function",
    argvalues=[
        SUB_FUNC,
        AND_FUNC,
        IAND_FUNC,
        OR_FUNC,
        IOR_FUNC,
        XOR_FUNC,
        IXOR_FUNC,
    ],
)
def test_no_segmentation_fault_intbitset_set(function):
    """intbitset - test no segmentation fault with foreign data types"""
    with pytest.raises(TypeError):  # NOQA
        function.intbitset_function(intbitset([1, 2, 3]), set([1, 2, 3]))


def test_no_segmentation_fault_intbitset_set_isub():
    """intbitset - test no segmentation fault with foreign data types"""
    ISUB_FUNC.intbitset_function(intbitset([1, 2, 3]), set([1, 2, 3]))


@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_no_segmentation_fault_set_intbitset(function):
    function.set_function(set([1, 2, 3]), intbitset([1, 2, 3]))


@pytest.mark.parametrize(
    argnames="function",
    argvalues=FUNCTIONS,
)
def test_no_segmentation_fault_none_intbitset(function):
    """intbitset - test no segmentation fault with foreign data types"""
    with pytest.raises(TypeError):  # NOQA
        function.intbitset_function(None, intbitset([1, 2, 3]))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_list_dump(set1):
    """intbitset - list dump"""
    assert set1 == list(intbitset(set1))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_ascii_bit_dump(set1):
    """intbitset - ascii bit dump"""
    tot = 0
    count = 0
    for bit in intbitset(set1).strbits():
        if bit == "0":
            assert count not in set1
        elif bit == "1":
            assert count in set1
            tot += 1
        else:
            raise Exception()
        count += 1
    assert len(set1) == tot


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_tuple_of_tuples(set1):
    """intbitset - support tuple of tuples"""
    tmp_tuple = tuple([(elem,) for elem in set1])
    assert list(intbitset(set1)) == list(intbitset(tmp_tuple))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_tuple_of_tuples_with_trailing_bits(set1):
    tmp_tuple = tuple([(elem,) for elem in set1])
    assert intbitset(set1, trailing_bits=True) == intbitset(tmp_tuple, trailing_bits=True)


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames="trailing_bits",
    argvalues=[True, False],
)
def test_marshalling(set1, trailing_bits):
    assert intbitset(set1, trailing_bits=trailing_bits) == intbitset(
        intbitset(set1, trailing_bits=trailing_bits).fastdump()
    )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames="trailing_bits",
    argvalues=[True, False],
)
def test_pickling(set1, trailing_bits):
    assert intbitset(set1, trailing_bits=trailing_bits) == pickle.loads(
        pickle.dumps(intbitset(set1, trailing_bits=trailing_bits), -1)
    )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_emptiness(set1):
    assert (not set(set1)) == (not intbitset(set1))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_len(set1):
    intbitset1 = intbitset(set1)
    pythonset1 = set(set1)
    assert len(pythonset1) == len(intbitset1)
    intbitset1.add(76543)
    pythonset1.add(76543)
    assert len(pythonset1) == len(intbitset1)
    intbitset1.remove(76543)
    pythonset1.remove(76543)
    assert len(pythonset1) == len(intbitset1)


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_clear(set1):
    intbitset1 = intbitset(set1)
    intbitset1.clear()
    assert list(intbitset1) == []
    intbitset1 = intbitset(set1, trailing_bits=True)
    intbitset1.clear()
    assert list(intbitset1) == []


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames="trailing_bits",
    argvalues=[True, False],
)
def test_set_repr(set1, trailing_bits):
    intbitset1 = intbitset(set1, trailing_bits=trailing_bits)
    assert eval(repr(intbitset1)) == intbitset1


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames="set2",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames=["intbitset_op", "set_op"],
    argvalues=[
        (intbitset.__eq__, set.__eq__),  # NOQA
        (intbitset.__ge__, set.__ge__),  # NOQA
        (intbitset.__gt__, set.__gt__),  # NOQA
        (intbitset.__le__, set.__le__),  # NOQA
        (intbitset.__lt__, set.__lt__),  # NOQA
        (intbitset.__ne__, set.__ne__),  # NOQA
    ],
)
def test_intbitset_behaves_same_as_set_cmp(set1, set2, intbitset_op, set_op):
    """intbitset - (non infinite) set comparison"""
    expected = set_op(set(set1), set(set2))
    result = intbitset_op(intbitset(set1), intbitset(set2))
    assert expected == result


@pytest.mark.parametrize(
    argnames="intbitset_op",
    argvalues=[
        intbitset.__eq__,  # NOQA
        intbitset.__ge__,  # NOQA
        intbitset.__gt__,  # NOQA
        intbitset.__le__,  # NOQA
        intbitset.__lt__,  # NOQA
        intbitset.__ne__,  # NOQA
    ],
)
def test_intbitset_richcmp_with_non_intbitset_return_false(intbitset_op):
    set1 = [1, 2, 3]
    assert not intbitset_op(intbitset(set1), set1)
    with pytest.raises(TypeError):  # NOQA
        intbitset_op(set1, intbitset(set1))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_update_with_signs(set1):
    dict1 = {10: -1, 20: 1, 23: -1, 27: 1, 33: -1, 56: 1, 70: -1, 74: 1}
    intbitset1 = intbitset(set1)
    intbitset1.update_with_signs(dict1)
    up_to = max(list(dict1.keys()) + set1)
    for i in range(up_to + 1):
        if dict1.get(i, i in set1 and 1 or -1) == 1:
            assert i in intbitset1, (
                "%s was not correctly updated from %s by %s"
                % (repr(intbitset1), repr(set1), repr(dict1)),
            )
        else:
            assert i not in intbitset1, (
                "%s was not correctly updated from %s by %s"
                % (repr(intbitset1), repr(set1), repr(dict1)),
            )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
@pytest.mark.parametrize(
    argnames="trailing_bits",
    argvalues=[True, False],
)
def test_set_cloning(set1, trailing_bits):
    intbitset1 = intbitset(set1, trailing_bits=trailing_bits)
    intbitset2 = intbitset(intbitset1)
    intbitset3 = copy.deepcopy(intbitset2)
    check_bitset(intbitset1)
    check_bitset(intbitset2)
    check_bitset(intbitset3)
    assert intbitset2 == intbitset1
    assert intbitset3 == intbitset1


def test_set_isdisjoint():
    sets = [
        intbitset(set([1, 2])),
        intbitset(set([3, 4])),
        intbitset(set([2, 3])),
    ]

    for set1 in sets:
        for set2 in sets:
            if set1 is not set2:
                assert set1.isdisjoint(set2) is set(set1).isdisjoint(set(set2))


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_pop(set1):
    intbitset1 = intbitset(set1)
    pythonlist1 = list(set1)
    while True:
        try:
            res1 = pythonlist1.pop()
        except IndexError:
            with pytest.raises(KeyError):  # NOQA
                intbitset1.pop()
            check_bitset(intbitset1)
            break
        res2 = intbitset1.pop()
        check_bitset(intbitset1)
        assert res2 == res1


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_getitem(set1):
    intbitset1 = intbitset(set1)
    pythonlist1 = list(set1)
    for i in range(-2 * len(set1) - 2, 2 * len(set1) + 2):
        try:
            res1 = pythonlist1[i]
        except IndexError:
            with pytest.raises(IndexError):  # NOQA
                intbitset1.__getitem__(i)
            continue

        assert intbitset1[i] == res1


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_getitem_slices(set1):
    intbitset1 = intbitset(set1)
    pythonlist1 = list(set1)
    for start in range(-2 * len(set1) - 2, 2 * len(set1) + 2):
        for stop in range(-2 * len(set1) - 2, 2 * len(set1) + 2):
            for step in range(1, 3):
                res1 = pythonlist1[start:stop:step]
                res2 = intbitset1[start:stop:step]
                assert list(res2) == res1, (
                    f"Failure: set={set1}, start={start}, stop={stop}, "
                    f"step={step}, found={list(res2)}, "
                    f"expected={res1}, "
                    + "indices: "
                    + str(slice(start, stop, step).indices(len(pythonlist1)))
                )


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_iterator(set1):
    intbitset1 = intbitset(set1)
    check_bitset(intbitset1)
    tmp_set1 = []
    for recid in intbitset1:
        check_bitset(intbitset1)
        tmp_set1.append(recid)
    check_bitset(intbitset1)
    assert tmp_set1 == set1


@pytest.mark.parametrize(
    argnames="set1",
    argvalues=TEST_SETS + [[]],
)
def test_set_iterator2(set1):
    tmp_set1 = []
    for recid in intbitset(set1):
        tmp_set1.append(recid)
    assert set1 == tmp_set1


@pytest.mark.parametrize(
    argnames="strdump",
    argvalues=[
        b("ciao"),
        b(intbitset([2, 6000000]).strbits()),
        b("djflsdkfjsdljfsldkfjsldjlfk"),
    ],
    ids=lambda s: s[:30],
)
def test_set_corruption(strdump):
    """intbitset - set corruption"""
    set1 = intbitset()

    # These should fail because they are not compressed
    with pytest.raises(ValueError):  # NOQA
        intbitset(strdump)

    with pytest.raises(ValueError):  # NOQA
        set1.fastload(strdump)

    # These should fail because they are not of the good length
    strdump = zlib.compress(strdump)

    with pytest.raises(ValueError):  # NOQA
        intbitset(strdump)

    with pytest.raises(ValueError):  # NOQA
        set1.fastload(strdump)


@pytest.mark.parametrize(
    argnames=["original", "dumped", "dumped_trailing_bits"],
    argvalues=[
        (
            (20, 30, 1000, 40),
            b("x\x9cc`\x10p``d\x18\x18\x80d/\x00*\xb6\x00S"),
            b("x\x9cc`\x10p`\x18(\xf0\x1f\x01\x00k\xe6\x0bF"),
        ),
        (
            (20, 30, 1000, 41),
            b("x\x9cc`\x10p``b\x18\x18\xc0\x88`\x02\x00+9\x00T"),
            b("x\x9cc`\x10p`\x18(\xf0\x1f\x01\x00k\xe6\x0bF"),
        ),
        (
            (20, 30, 1001, 41),
            b("x\x9cc`\x10p``b\x18\x18\x80d/\x00+D\x00U"),
            b("x\x9cc`\x10p`\x18(\xf0\xef?\x1c\x00\x00k\xdb\x0bE"),
        ),
    ],
    ids=lambda s: s[0],
)
def test_set_consistency(original, dumped, dumped_trailing_bits):
    intbitset1 = intbitset(original)
    intbitset2 = intbitset(original, trailing_bits=True)
    intbitset3 = intbitset(dumped)
    intbitset4 = intbitset(dumped_trailing_bits)
    check_bitset(intbitset1)
    check_bitset(intbitset2)
    check_bitset(intbitset3)
    check_bitset(intbitset4)

    assert intbitset1.fastdump() == dumped
    assert intbitset1 == intbitset3
    assert intbitset2.fastdump() == dumped_trailing_bits
    assert intbitset2 == intbitset4


def test_empty_generator():
    intbitset(range(0))
    intbitset(i for i in range(0))


def test_do_not_allow_removal_of_none():
    with pytest.raises(TypeError):  # NOQA
        intbitset([1, 2, 3]) - None



# -*- coding: utf-8 -*-
#
# This file is part of intbitset
# Copyright (C) CERN and others
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# intbitset is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# intbitset is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with intbitset; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""C-based extension implementing fast integer bit sets."""

from setuptools import Extension
from setuptools import setup

setup(
    name="intbitset",
    version="4.0.0",
    url="http://github.com/inveniosoftware-contrib/intbitset/",
    license="LGPL-3.0-or-later",
    author="Invenio collaboration, maintained by Philippe Ombredanne and AboutCode.org",
    author_email="pombredanne@nexb.com",
    description="C-based extension implementing fast integer bit sets.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    package_dir={"": "intbitset"},
    py_modules=["intbitset_helper", "intbitset_version"],
    ext_modules=[
        Extension(
            "intbitset",
            ["intbitset/intbitset.c", "intbitset/intbitset_impl.c"],
            # extra_compile_args = []
            # For debug -> extra_compile_args=['-ftree-vectorizer-verbose=2']
        )
    ],
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Cython",
        "Programming Language :: C",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
    ],
    extras_require={"tests": ["pytest", "pytest-xdist"]}
)



# This file is part of Invenio.
# Copyright (C) 2007, 2008, 2009, 2010, 2011, 2013, 2015, 2016 CERN.
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""
This function is needed only when load a pickled instanced of intbitset.
It needs to be stored in a real .py file, because currently cython
would have instantiate such a function as built-in function.
"""


def _(dump):
    """
    As part of the pickle protocol, a callable that will instantiate
    the class is needed.
    It's called _ just to make it short and hidden :-)
    """
    from intbitset import intbitset

    return intbitset(dump)



# -*- coding: utf-8 -*-
#
# This file is part of intbitset
# Copyright (C) 2015, 2016 CERN.
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
# intbitset is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# intbitset is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with intbitset; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Version information for intbitset."""

# Do not change the format of this next line. Doing so risks breaking
# setup.py and docs/conf.py

__version__ = "3.1.0"
